import time
from blinker import signal
from olm.helper import Map
from olm.logger import get_logger

logger = get_logger('olm.signal')

signals = Map({
    'INITIALISED':              "INITIALISED",             # args: context
    'SITE_INITIALISED':         "SITE_INITIALISED",        # args: context
    "BEFORE_MD_CONVERT":        "BEFORE_MD_CONVERT",       # args: context, content      
    'BEFORE_CACHING':           "BEFORE_CACHING",          # args: context, articles
    'AFTER_ARTICLE_READ':       "AFTER_ARTICLE_READ",      # args: context, article
    'AFTER_PAGE_READ':          "AFTER_PAGE_READ",         # args: context, article
    'AFTER_ALL_ARTICLES_READ':  "AFTER_ALL_ARTICLES_READ", # args: context, articles
    'BEFORE_WRITING':           "BEFORE_WRITING",          # args: context
    'BEFORE_ARTICLE_WRITE':     "BEFORE_ARTICLE_WRITE",    # args: context, article
    'AFTER_WRITING':            "AFTER_WRITING"            # args: context, article
})

class Signal:
    def __init__(self, signal_value):
        logger.spam('Instantiating new Signal "%s"', signal_value)
        if signal_value not in signals:
            logger.error('No such signal %s', signal_value)
            raise Exception('No such signal %s', signal_value)
        self.signal_value = signal_value
        self.sender = signal(signal_value)

    def send(self, **kwargs):
        time_start = time.time()
        logger.spam('Sending signal "%s"', self.signal_value)
        self.sender.send(self.signal_value, **kwargs)
        logger.spam('Ran signal "%s" in %.3f seconds', self.signal_value, (time.time() - time_start))




import os
import sys
import imp
import mistune
import re
from collections import OrderedDict
from olm.helper import Map, merge_ordered_dictionaries
from olm.logger import get_logger
from jinja2 import Environment, FileSystemLoader, select_autoescape

logger = get_logger('olm.context')

def load_default_context(path):
    logger.info('Loading default context')
    BASE_FOLDER          = os.path.abspath(path)
    CACHE_LOCATION       = os.path.abspath(os.path.join(BASE_FOLDER)) + os.sep + 'cache.pickle'
    SOURCE_FOLDER        = os.path.abspath(os.path.join(BASE_FOLDER, 'src'))
    STATIC_FOLDER        = os.path.abspath(os.path.join(BASE_FOLDER, 'theme', 'static'))
    CSS_FOLDER           = os.path.abspath(os.path.join(BASE_FOLDER, 'theme', 'static', 'css'))
    JS_FOLDER            = os.path.abspath(os.path.join(BASE_FOLDER, 'theme', 'static', 'js'))
    TEMPLATES_FOLDER     = os.path.abspath(os.path.join(BASE_FOLDER, 'theme', 'templates'))
    PLUGINS_FOLDER       = os.path.abspath(os.path.join(BASE_FOLDER, 'plugins',))
    MD                   = mistune.Markdown(escape=False)
    JINJA_ENV            = Environment(
                            loader=FileSystemLoader([TEMPLATES_FOLDER])
                            )
    ARTICLE_TYPES        = ['article']
    INDEX_TYPES          = ['index']
    PLUGINS              = []
    SITEURL              = ""
    OUTPUT_FOLDER        = os.path.abspath(os.path.join(BASE_FOLDER, 'dist'))
    OUTPUT_CSS_FOLDER    = os.path.abspath(os.path.join(OUTPUT_FOLDER, 'theme', 'css'))
    OUTPUT_JS_FOLDER     = os.path.abspath(os.path.join(OUTPUT_FOLDER, 'theme', 'js'))
    NO_SCAN              = []
    ARTICLE_WRITE_TRIGGERS      = []
    ARTICLE_META_WRITE_TRIGGERS = []
    PAGE_WRITE_TRIGGERS         = []
    PAGE_META_WRITE_TRIGGERS    = []
    INDEX_WRITE_TRIGGERS        = ["ARTICLE.NEW_FILE"]
    INDEX_META_WRITE_TRIGGERS   = []

    CONTEXT = OrderedDict({
        "BASE_FOLDER":          BASE_FOLDER,
        "CACHE_LOCATION":       CACHE_LOCATION,
        "SOURCE_FOLDER":        SOURCE_FOLDER,
        "OUTPUT_FOLDER":        OUTPUT_FOLDER,
        "STATIC_FOLDER":        STATIC_FOLDER,
        "CSS_FOLDER":           CSS_FOLDER,
        "JS_FOLDER":            JS_FOLDER,
        "MD":                   MD,
        "JINJA_ENV":            JINJA_ENV,
        "ARTICLE_TYPES":        ARTICLE_TYPES,
        "INDEX_TYPES":          INDEX_TYPES,
        "PLUGINS":              PLUGINS,
        "PLUGINS_FOLDER":       PLUGINS_FOLDER,
        "SITEURL":              SITEURL,
        "SUBSITES":             {},
        "OUTPUT_CSS_FOLDER":    OUTPUT_CSS_FOLDER,
        "OUTPUT_JS_FOLDER":     OUTPUT_JS_FOLDER,
        "NO_SCAN":              NO_SCAN,
        "ARTICLE_WRITE_TRIGGERS":      ARTICLE_WRITE_TRIGGERS,
        "ARTICLE_META_WRITE_TRIGGERS": ARTICLE_META_WRITE_TRIGGERS,
        "PAGE_WRITE_TRIGGERS":         PAGE_WRITE_TRIGGERS,
        "PAGE_META_WRITE_TRIGGERS":    PAGE_META_WRITE_TRIGGERS,
        "INDEX_WRITE_TRIGGERS":        INDEX_WRITE_TRIGGERS,
        "INDEX_META_WRITE_TRIGGERS":   INDEX_META_WRITE_TRIGGERS,
    })

    return CONTEXT

def load_context_from_file(settings_file_path, prev_context=None):
    sys_path = sys.path
    sys.path.append(os.path.dirname(settings_file_path))
    py_mod = imp.load_source('settings', settings_file_path)
    user_settings = getattr(py_mod, 'SETTINGS')
    sys.path = sys_path

    return load_context(user_settings, prev_context)


def load_context(context, prev_context=None):
    logger.info('Loading site context')
    
    # Combine default and site settings
    if prev_context is not None:
        CONTEXT = merge_ordered_dictionaries(prev_context, context)
    else:
        CONTEXT = context

    replace_happened = True
    while replace_happened:
        replace_happened = False
        for key in CONTEXT:
            # Do variable substitutions in settings strings
            if isinstance(CONTEXT[key], str):
                value = CONTEXT[key]
                pattern = re.compile(r'{{\s*(\S*)\s*}}')
                for match in re.finditer(pattern, CONTEXT[key]):
                    try:
                        value = value.replace(match.group(0), CONTEXT[match.group(1)])
                        replace_happened = True
                    except KeyError as e:
                        logger.warn("Can't replace '%s' in settings file for key %s", match.group(1), key)
                        logger.warn(e)
                CONTEXT[key] = value

            # Reset the Jinja Env if we have a new templates folder
            if key == "TEMPLATES_FOLDER":
                CONTEXT["JINJA_ENV"] = Environment(
                    loader=FileSystemLoader([CONTEXT[key]])
                )
        
    return Map(CONTEXT)



import codecs
import os
from olm.helper import merge_dictionaries

class Writer:
    def __init__(self, context, relative_path, template, **kwargs):
        self.context = context
        self.relative_path = relative_path
        self.template = context.JINJA_ENV.get_template(template)
        self.kwargs = kwargs

        self.output_filepath = os.path.join(context.OUTPUT_FOLDER, relative_path)


    def write_file(self):
        """Write the article to a file"""
        os.makedirs(os.path.dirname(self.output_filepath), exist_ok=True)
        with codecs.open(self.output_filepath, 'w', encoding='utf-8') as html_file:
            html = self.template.render(**merge_dictionaries(self.context, self.kwargs))
            html_file.write(html)


import logging, verboselogs, coloredlogs

LEVEL = None

def set_log_level(level='INFO'):
    global LEVEL
    LEVEL=level

def get_logger(name=''):
    logger = verboselogs.VerboseLogger(name)
    coloredlogs.install(level=LEVEL, logger=logger, fmt='%(asctime)s [%(name)s] %(message)s')
    return logger


class ArticleStatus:
    ACTIVE = "active"
    UNLISTED = "unlisted"
    DRAFT = "draft"

class CacheTypes:
    NEW_FILE = "NEW_FILE"
    META_CHANGE = "META_CHANGE"
    CONTENT_CHANGE = "CONTENT_CHANGE"
    REMOVED_FILE = "REMOVED_FILE"


import pickle
import json
import time
import os
import hashlib
from olm.logger import get_logger
from olm.constants import CacheTypes, ArticleStatus
from collections import namedtuple

logger = get_logger('olm.cache')
comparison = namedtuple('comparison', ['added', 'removed', 'modified'])

def dict_compare(d1, d2):
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    intersect_keys = d1_keys.intersection(d2_keys)
    added_keys = d1_keys - d2_keys
    removed_keys = d2_keys - d1_keys
    modified = {o: (d1[o], d2[o]) for o in intersect_keys if d1[o] != d2[o]}
    same = set(o for o in intersect_keys if d1[o] == d2[o])
    added = {}
    removed = {}
    for key in added_keys:
        added[key] = d1[key]
    for key in removed_keys:
        removed[key] = d2[key]
    return comparison(added=added, removed=removed, modified=modified)

def hash_object(thing):
    return hashlib.md5(json.dumps(thing, sort_keys=True).encode('utf-8')).hexdigest()

def add_change(file_status, change_list, file_type, change_type):
    change = '{}.{}'.format(file_type, change_type)
    if change not in change_list:
        change_list.append(change)
    return change_list

def check_cache(CONTEXT, files):
    time_start = time.time()
    CONTEXT['cache_change_types'] = []
    CONTEXT['cache_changed_meta'] = []
    if not CONTEXT.caching_enabled:
        logger.info('Caching disabled')
        return
    
    # Load the cache file
    old_hashes = {}
    CONTEXT['is_cached'] = False
    if os.path.isfile(CONTEXT.CACHE_LOCATION):
        with open(CONTEXT.CACHE_LOCATION, 'rb') as handle:
            old_hashes = pickle.load(handle)
            CONTEXT['is_cached'] = True

    hashes       = {} # New hashes to be saved to cache file
    changes      = [] # List of change types
    meta_changes = [] # List of all changed metadata
    no_of_files  = 0
    for afile in files:
        if 'cache_id' not in dir(afile):
            continue
        # Get hashes of id and content
        m               = hashlib.md5()
        id_hash         = hash_object(afile.cache_id)
        meta_hash       = hash_object(afile.metadata)
        content_hash    = hash_object(afile.content)
        hashes[id_hash] = (meta_hash, content_hash, afile.metadata, afile.cache_type)

        # Figure out an identifier for logging
        if afile.output_filepath is None:
            identifier = afile.basename
        else:
            identifier = afile.output_filepath
        
        # If the id hash exists in the old hashes check for changes
        # else add new file change
        if id_hash in old_hashes:
            # If difference is in metadata compile a diff object for the metadata
            # and store it along with a meta change
            if old_hashes[id_hash][0] != meta_hash:
                logger.debug('{} metadata is different to cache'.format(identifier))
                logger.debug('{}'.format(old_hashes[id_hash][2]))
                logger.debug('{}'.format(afile.metadata))
                diff = dict_compare(old_hashes[id_hash][2], afile.metadata)
                meta_changes.append(diff)
                changes = add_change(afile.status, changes, afile.cache_type, CacheTypes.META_CHANGE)
                no_of_files = no_of_files + 1
            # If difference in content then just add a content change
            if old_hashes[id_hash][1] != content_hash:
                logger.debug('{} content is different to cache'.format(identifier))
                changes = add_change(afile.status, changes, afile.cache_type, CacheTypes.CONTENT_CHANGE)
                no_of_files = no_of_files + 1
            # If same mark the file as same_as_cache so it doesnt get written
            if old_hashes[id_hash][0] == meta_hash and old_hashes[id_hash][1] == content_hash:
                afile.same_as_cache = True
        else:
            logger.debug('{} is a new file'.format(identifier))
            changes = add_change(afile.status, changes, afile.cache_type, CacheTypes.NEW_FILE)
            meta_changes.append(comparison(added=afile.metadata, removed={}, modified={}))
            no_of_files = no_of_files + 1

    removed = list(set(old_hashes.keys()) - set(hashes.keys()))
    for r in removed:
        changes = add_change(None, changes, old_hashes[r][3], CacheTypes.REMOVED_FILE)
        meta_changes.append(comparison(added={}, removed=old_hashes[r][2], modified={}))

    logger.info('{} files are new or changed'.format(no_of_files))
    logger.debug('{} are the new changes'.format(changes))
    CONTEXT['cache_change_types'] = changes
    CONTEXT['cache_changed_meta'] = meta_changes
   
    # Write the latest build's caches to file
    with open(CONTEXT.CACHE_LOCATION, 'wb') as handle:
        pickle.dump(hashes, handle, protocol=pickle.HIGHEST_PROTOCOL)

    CONTEXT['cache'] = hashes
    logger.info('Cache finished in %.3f', time.time() - time_start)


import os
import codecs
from olm.reader import Reader
from olm.writer import Writer
from olm.signals import Signal, signals

class Source:
    """ Represents any markdown source file in the site e.g. an Article or a Page """
    def __init__(self, context, filepath=None, metadata=None, content=None, basename=None):
        self.context = context
        if filepath is not None:
            self.source_filepath = filepath
            # Get filenames, paths etc
            self.dirname = os.path.dirname(filepath)
            self.basepath, self.filename = os.path.split(filepath)
            self.basename, self.extension = os.path.splitext(self.filename)
            self.relpath = os.path.relpath(os.path.join(self.dirname, self.basename) + '.html', context.SOURCE_FOLDER)
            
            # Parse the file for content and metadata
            with codecs.open(filepath, 'r', encoding='utf8') as md_file:
                reader = Reader(md_file.read())
                metadata, content = reader.parse()

        elif metadata is not None and content is not None and basename is not None:
            self.content = content
            self.basename = basename
            self.source_filepath = None
            self.dirname  = None
            self.basepath = None
            self.relpath  = None

        else:
            raise Exception('Article object not supplied with either filepath or content and metadata.') 
        
        #TODO: this doesnt seem to work
        #signal_sender = Signal(signals.BEFORE_MD_CONVERT)
        #signal_sender.send(context=context, content=raw_content)
        self.content = content
        self.metadata = metadata
        self.status = None
        self.template = None
        self.output_filepath = None
        self.same_as_cache = False
        self.cache_type = 'SOURCE'

    def write_file(self, context=None, **kwargs):
        self.context = context if context is not None else self.context
        if self.context.caching_enabled and self.same_as_cache:
            return
        if self.template is None:
            return
        self.content = context.MD(self.content)
        writer = Writer(
            self.context, 
            self.output_filepath, 
            self.template,
            **kwargs)
        writer.write_file()


import sys
import os
import time
import codecs
import sass
from jsmin import jsmin
from shutil import copyfile

from olm.context import load_context, load_default_context
from olm.plugins import Plugins
from olm.article import Article
from olm.index import Index
from olm.page import Page
from olm.constants import ArticleStatus
from olm.helper import Map
from olm.writer import Writer
from olm.signals import Signal, signals
from olm.logger import get_logger
from olm.cache import check_cache

logger = get_logger('olm.site')

class Site:
    def __init__(self, context):
        self.context = context
        self.subsites = []
    
    def build_site(self):
        CONTEXT = self.context
        all_files = []
        articles = []
        draft_articles = []
        unlisted_articles = []
        pages = []
        subsites = set()

        CONTEXT['authors'] = {}
        CONTEXT['all_files'] = []
        CONTEXT['articles'] = []
        CONTEXT['pages'] = []

        signal_sender = Signal(signals.SITE_INITIALISED)
        signal_sender.send(context=CONTEXT)

        # Scan source files
        logger.info("Scanning source files")
        time_source_start = time.time()
        for dirname, dirs, files in os.walk(CONTEXT.SOURCE_FOLDER):
            for filename in files:
                filepath = os.path.join(dirname, filename)
                relpath = os.path.relpath(filepath, CONTEXT.SOURCE_FOLDER)
                firstfolder = relpath.split(os.sep)[0]
                basename, extension = os.path.splitext(filename)
                if firstfolder in CONTEXT.NO_SCAN:
                    continue
                if extension.lower() == ".md":
                    if firstfolder[0] == "_":
                        subsites.add(firstfolder)
                    elif firstfolder == "pages":
                        logger.debug("Found %s", filepath)
                        page = Page(CONTEXT, filepath)
                        pages.append(page)
                        all_files.append(page)
                    else:
                        logger.debug("Found %s", filepath)
                        article = Article(CONTEXT, filepath)
                        if article.type in CONTEXT.ARTICLE_TYPES + CONTEXT.INDEX_TYPES:
                            if article.status == ArticleStatus.ACTIVE:
                                articles.append(article)
                            elif article.status == ArticleStatus.UNLISTED:
                                unlisted_articles.append(article)
                            else:
                                draft_articles.append(article)
                        all_files.append(article)
        logger.info("Processed %d articles, %d unlisted articles, %d drafts, and %d pages in %.3f seconds", len(articles), len(unlisted_articles), len(draft_articles), len(pages), time.time() - time_source_start)            

        # Extend the lists (in case a plugin has added to it post init)
        CONTEXT['all_files'].extend(all_files)
        CONTEXT['articles'].extend(sorted(articles, key=lambda k: (k.date), reverse=True))
        CONTEXT['pages'].extend(pages)

        signal_sender = Signal(signals.BEFORE_CACHING)
        signal_sender.send(context=CONTEXT, articles=CONTEXT.articles)

        # Work out the cache status of all the fles
        check_cache(CONTEXT, CONTEXT['all_files'])
        
        signal_sender = Signal(signals.AFTER_ALL_ARTICLES_READ)
        signal_sender.send(context=CONTEXT, articles=CONTEXT.articles)

        # Check for duplicate output paths
        # and see if output file exists
        outputs = []
        for f in CONTEXT['all_files']:
            if f.output_filepath is None:
                continue
            if f.output_filepath not in outputs:
                outputs.append(f.output_filepath)
                if not os.path.isfile(f.output_filepath):
                    logger.spam("'%s' does not exist. Setting cached status to False.", f.source_filepath)
                    #f.same_as_cache = False
            else: 
                dupes = [ b for b in articles if b.output_filepath == f.output_filepath ]
                logger.error("'%s' has the same output file path as '%s'. The other file will be overwritten.", f.source_filepath, dupes[0].source_filepath)

        signal_sender = Signal(signals.BEFORE_WRITING)
        signal_sender.send(context=CONTEXT, Writer=Writer)

        # Write all the articles
        all_articles = CONTEXT.articles + unlisted_articles
        logger.debug("Writing %d articles", len(all_articles))
        time_write_start = time.time()
        number_written = 0
        for index, article in enumerate(all_articles):
            logger.spam("Writing file %d of %d", index + 1, len(all_articles))
            wrote = article.write_file(context=CONTEXT)
            number_written = number_written + 1 if wrote else number_written
        logger.info("Wrote %d changed articles out of %d articles in %.3f seconds", number_written, len(all_articles), (time.time() - time_write_start))

        # Write all the pages
        logger.debug("Writing %d pages", len(pages))
        time_write_start = time.time()
        number_written = 0
        for index, page in enumerate(pages):
            logger.spam("Writing file %d of %d", index + 1, len(pages))
            wrote = page.write_file(context=CONTEXT)
            number_written = number_written + 1 if wrote else number_written
        logger.info("Wrote %d changed pages out of %d pages in %.3f seconds", number_written, len(pages), (time.time() - time_write_start))

        # Write the index
        logger.debug("Writing articles index")
        time_write_start = time.time()
        index = Index(CONTEXT)
        wrote = index.write_file()
        if wrote:
            logger.info("Wrote index in %.3f seconds", (time.time() - time_write_start))
        else:
            logger.info("Reused cached index in %.3f seconds", (time.time() - time_write_start))

        signal_sender = Signal(signals.AFTER_WRITING)
        signal_sender.send(context=CONTEXT, Writer=Writer)

        # Find, compile and move static files
        logger.debug("Compiling static files")
        time_static_start = time.time()
        # SASS
        sass.compile(dirname=(CONTEXT.CSS_FOLDER, CONTEXT.OUTPUT_CSS_FOLDER), output_style='compressed')
        # CSS
        for dirname, dirs, files in os.walk(CONTEXT.CSS_FOLDER):
            for filename in files:
                filepath = os.path.join(dirname, filename)
                basename, extension = os.path.splitext(filename)
                rel_path = os.path.relpath(filepath, CONTEXT.CSS_FOLDER)
                if extension.lower() == ".css":
                    os.makedirs(os.path.dirname(os.path.join(CONTEXT.OUTPUT_CSS_FOLDER, rel_path)), exist_ok=True)
                    copyfile(filepath, os.path.join(CONTEXT.OUTPUT_CSS_FOLDER, rel_path))
        # JS
        for dirname, dirs, files in os.walk(CONTEXT.JS_FOLDER):
            for filename in files:
                filepath = os.path.join(dirname, filename)
                basename, extension = os.path.splitext(filename)
                rel_path = os.path.relpath(filepath, CONTEXT.JS_FOLDER)
                if extension.lower() == ".js":
                    with codecs.open(filepath, encoding='utf-8', errors='ignore') as js_file:
                        minified = js_file.read()
                    output_filepath = os.path.join(CONTEXT.OUTPUT_JS_FOLDER, rel_path)
                    os.makedirs(os.path.dirname(output_filepath), exist_ok=True)
                    with codecs.open(output_filepath, 'w+', encoding='utf-8') as js_min_file:
                        js_min_file.write(minified)
        logger.info("Processed static files in %.3f seconds", time.time() - time_static_start)
        
        self.subsites = subsites


import os
from os.path import dirname, splitext, relpath, join
import datetime
import codecs
import re

from olm.signals import signals, Signal
from olm.source import Source
from olm.helper import merge_dictionaries

class Page(Source):
    """Object representing an article"""

    def __init__(self, context, filepath=None):
        super().__init__(context, filepath)

        self.template        = 'page.html'
        self.title           = self.metadata['title'] if 'title' in self.metadata else self.basename
        self.url             = self.relpath

        if 'PAGE_SLUG' in context:
            slug_dict = vars(self)
            output_filename = context.PAGE_SLUG.format(**slug_dict)
        else:
            output_filename = '{}.html'.format(self.basename)
        
        self.output_filepath = os.path.join(context.OUTPUT_FOLDER, 'pages', output_filename)
        self.url             = 'pages/{}'.format(output_filename)

        self.cache_id = self.output_filepath
        self.cache_type = 'PAGE'
        
        signal_sender = Signal(signals.AFTER_PAGE_READ)
        signal_sender.send(context=context, page=self)


    def write_file(self, context=None):
        changes                = self.context['cache_change_types']
        changed_meta           = self.context['cache_changed_meta']
        refresh_triggers       = self.context['PAGE_WRITE_TRIGGERS']
        refresh_meta_triggers  = self.context['PAGE_META_WRITE_TRIGGERS']
        if any(i in changes for i in refresh_triggers):
            self.same_as_cache = False
        if any(any(m in merge_dictionaries(*c) for m in refresh_meta_triggers) for c in changed_meta):
            self.same_as_cache = False
        super().write_file(context, page=self)
        return not self.same_as_cache


import re

INDENTATION = re.compile(r'^(\s{4,}|\t{2,})(\S.*)')
META = re.compile(r'^(\w+):\s*(.*)')

class Reader:
    def __init__(self, text):
        self.text = text
        self.metadata = {}
        self.content = ""
        self.keys = []

    def parse(self):
        """Parse the given text into metadata and strip it for a Markdown parser.
        :param text: text to be parsed
        """
        text = self.text.split('\n')
        meta = {}
        m = META.match(text[0])

        while m:
            key = m.group(1)
            value = m.group(2)
            meta[key] = value
            text = text[1:]
            if len(text) >= 1:
                i = INDENTATION.match(text[0])
                while i:
                    if not isinstance(meta[key], list):
                        meta[key] = [meta[key]]
                    meta[key].append(i.group(2))
                    text = text[1:]
                    i = INDENTATION.match(text[0])
                m = META.match(text[0])
            else:
                m = False

        for key in meta:
            self.metadata[key.lower()] = meta[key].strip() if isinstance(meta[key], str) else meta[key]
        content = '\n'.join(text)
        #TODO: make TOC optional?
        self.content = self.generate_toc(content)

        return self.metadata, self.content 

    def get_id(self, title):
        link_id = title
        index = 0
        while link_id in self.keys:
            index = index + 1
            link_id = title + '_' + str(index)
        self.keys.append(link_id)
        return link_id

    def generate_toc(self, content):
        pattern = re.compile(r'\[TOC\]')
        # Only run if [TOC] in text
        if pattern.search(content):
            replacements = []     
            lines =  content.split('\n')
            min = None
            headers = []
            hashpattern = re.compile(r'(#+)(.*)')
            # Find all headers and their place in the text, and
            # the minimum header number
            for index, line in enumerate(lines):
                if line.startswith('#'):
                    match = hashpattern.match(line)
                    no_of_hashes = len(match.group(1))
                    title = match.group(2).strip()
                    if min is None or no_of_hashes < min:
                        min = no_of_hashes
                    headers.append((index, no_of_hashes, title))
            output = ''
            prev_depth = 0
            for header in headers:
                index = header[0]
                no_of_hashes = header[1]
                depth = no_of_hashes - min + 1
                title = header[2]
                # Get unique id for header link
                link_id = self.get_id(title)
                # Replace markdown header with html header + id
                html_header = '<h{no_of_hashes} id="{link_id}">{title}</h{no_of_hashes}>\n'.format(title=title, link_id=link_id, no_of_hashes=no_of_hashes)
                replacements.append((index, html_header))
                # Construct list of header links by comparing
                # adjacent header numbers as list depth
                if prev_depth < depth:
                    for i in range(prev_depth, depth):
                        output = output + '<ul>\n'
                    output = output + '<li><a href="#{}">{}</a></li>\n'.format(link_id, title)
                elif prev_depth == depth:
                    output = output + '<li><a href="#{}">{}</a></li>\n'.format(link_id, title)
                elif prev_depth > depth:
                    for i in range(depth, prev_depth):
                        output = output + '</ul>\n'
                    output = output + '<li><a href="#{}">{}</a></li>\n'.format(link_id, title)
                prev_depth = depth  
            for i in range(0, prev_depth):
                output = output + '</ul>\n'
            # First replace all markdown headers so that the line
            # numbers are correct
            for replacement in replacements:
                lines[replacement[0]] = replacement[1]
            # Then replace [TOC] with the generated TOC
            return pattern.sub(output, '\n'.join(lines))
        else:
            return content